***********************************************************************
*     Template file for Laboration 3.                                 *
*                                                                     *
*     Some Fortran remarks: all statements must begin at column 7 and *
*     no line may exceed the 72th column. Exceptions: comments must   *
*     start with a char at column 1. To continue a statements on the  *
*     next line use a "$" token in the 6th column. Use capitals for   *
*     the statements for clarity.                                     *
*                                                                     *
***********************************************************************
*
*     Program declaration
*
      PROGRAM LABPROGRAM
*
*     Turning off implicit naming of variables - now all variables used
*     must be explicitely declared with name and type.
*
      IMPLICIT NONE
*
*     Declaration of used functions, varables etc.
*     
*     .. External Functions and Subroutines .. 
*
      INTEGER BLACS_PNUM, NUMROC, INDXG2P, CEIL
      DOUBLE PRECISION MPI_WTIME
      EXTERNAL BLACS_PNUM, NUMROC, MPI_WTIME,
     $         PDTRSM, PDORMQR, PDGEQRF, PDLACPY, PDGEMM
*
*     .. Internal Functions and Subroutines
*      
      INTRINSIC MAX
*
*     .. Scalar Variable Declaration ..
*
      DOUBLE PRECISION START_TIME, END_TIME, BNORM
      INTEGER CONTXT, IAM, NPROCS, NPROW, NPCOL, MYROW, MYCOL,
     $        AROWS, ACOLS, BCOLS, ICALLER, I, J,
     $        II, JJ, ICSRC, INFO, LDA, LDB, M, MB, N, NB, K, KB, BN, 
     $        BM, LWORK, NIN, LWORK_NEEDED, LTAUA_NEEDED, LTAUA,
     $        A, B, C, A_COPY, B_COPY, MEMSZ, DPMEMSZ, DPSZ, WORK, TAUA
*
*     .. Scalar Parameters Declaration ..
*
      PARAMETER (MEMSZ = 450000000, DPSZ = 8, DPMEMSZ = MEMSZ/DPSZ)
*
*     .. Arrays Variable Declaration .. 
*     
      INTEGER     DESCA(9), DESCB(9) 
      INTEGER     INP(4)
*
*     Here we declare the memory for our matrices A, B and copies of 
*     them as well as some workspace.
*
      DOUBLE PRECISION MEM( DPMEMSZ )
* 
      CHARACTER*40     INFILE
*     
      PARAMETER (NIN=11, INFILE='inp.dat')
*
*     GRID SETUP
*     Determine my process number (IAM) and the number of processes used in
*     machine (NPROCS) - the initialization of BLACS
*
*     TODO
      
*
*     Compute the number of rows and columns in the logical mesh
*     of processors
*
*     TODO
      NPROW = 
      NPCOL = 

*
*     Get default system context, and define grid as a mesh with the
*     configuration NPROW x NPCOL (find and assign the proper values
*     for CONTXT, MYCOL, MYROW)
*
*     TODO

*      
*     If some process was excluded from the grid, go to end of program
*     
      IF ( (MYROW.GE.NPROW ) .OR. (MYCOL.GE.NPCOL  ) ) GOTO 500
*
*     READING INFORMATION FROM THE FILE
*     Get my process ID (system ID) from my grid coordinates
*
      ICALLER = BLACS_PNUM( CONTXT, MYROW, MYCOL )
*      
*     If I am the process with the system number 0 read the input
*     file inp.dat from the working directory
*      
      IF( ICALLER.EQ.0 ) THEN
*         
*     Open the file
*
         OPEN( UNIT=NIN, FILE=INFILE, STATUS='OLD' )
*           
*     Read four integers from the file specifying the three
*     matrix dimensions (M,N,K) and a block size (NB)
*           
         READ( NIN, FMT=* ) ( INP(I), I=1,4 )
*           
*     Close the file
*
         CLOSE( NIN )
*           
*     Distribute the information to the processors in the current 
*     CONTXT (remember, a CONTEXT in BLACS = MPI COMMUNICATOR)
*
*     TODO
            
*        
*     All the other processes receive the integers
*
      ELSE
*     TODO        
           
      END IF
*
*     All processes assign the information received to the 
*     appropriate variables: M is #rows in A and B, N is #cols
*     in A and K is #cols in B. MB,NB,KB are blocking factors
*
      M  = INP(1)
      N  = INP(2)
      K  = INP(3)
      MB = INP(4)
      NB = INP(4)
      KB = INP(4)
*
*     "DISTRIBUTING" THE MATRICES A, B, A_COPY, B_COPY
*
*     Compute number of rows of A, cols of A and cols of B that belong
*     to my process
*
*     TODO
      AROWS = 
      ACOLS = 
      BCOLS = 

*
*     Then init descriptors DESCA and DESCB for the matrices A and B
*
*     TODO        

*
*     GENERATE THE MATRICES A AND B       
*     All processes generate their part of the distributed random matrices
*     A and B. 
*     First we divide the needed memory between the different arrays
*     by using ScaLAPACK "pointers"
*
      A        = 1
      B        = A      + AROWS * ACOLS
      A_COPY   = B      + AROWS * BCOLS
      B_COPY   = A_COPY + AROWS * ACOLS
*     Temporary workspace pointers (size 1)      
      TAUA     = B_COPY + AROWS * BCOLS
      WORK     = TAUA   + 1
*
*     Generate the local parts of the global matrices A and B using 
*     random generator
*
      CALL MATINIT( CONTXT, MEM(A), DESCA, MEM(B), DESCB, AROWS, 
     $              ACOLS, BCOLS)
*
*     COPYING THE MATRICES
*     Make a copy A and B to A_COPY and B_COPY for checking purposes
*      
*     TODO

*
*     SOLUTION PROCESS
*     Write some information to standard output
*
      IF( ICALLER.EQ.0 ) THEN
         WRITE(*,*) 'Solve the overdetermined system AX=B'
         WRITE(*,'(a,i0,a,i0)') 
     $      'Number of processes is ', NPROW,'x', NPCOL
         WRITE(*,'(a,i0,a,i0,a,i0,a,i0)') 
     $      'The matrices A and B are of the sizes ', M,'x',
     $      N,' and ',M,'x',K
      END IF
*
*     Check workspace by making calls to PDGEQRF and PDORMQR and
*     specifying -1 as the workspace size. Optimal workspace size
*     is then returned in first element in workspace.
*     Specify the other arguments exactly as you do later.
*
*     TODO (Check for PDGEQRF)
      LWORK_NEEDED = INT( MEM(WORK) )
      LTAUA_NEEDED = MAX( AROWS, ACOLS )
*     TODO (Check for PDORMQR)      
      LWORK_NEEDED = MAX( LWORK_NEEDED, INT( MEM(WORK) ))
*
*     Allocate memory for TAUA and WORK
*     
      LTAUA = MIN(LTAUA_NEEDED, DPMEMSZ - TAUA + 1)
      WORK  = TAUA + LTAUA
      LWORK = MIN(LWORK_NEEDED, DPMEMSZ - WORK + 1)
*
      IF( LWORK.LT.LWORK_NEEDED .OR. LTAUA.LT.LTAUA_NEEDED) THEN
         WRITE(*,*) 'Unable to solve the problem!'
         WRITE(*,*) 'REASON: lack of workspace.'
         GOTO 500
      END IF
*
*     Synchronize the processors and start the clock
*     
*     TODO

*
*     QR-factorize the matrix A by calling the ScaLAPACK routine PDGEQRF 
*
*     TODO     

*
*     Synchronize the processors and stop the clock
*     
*     TODO
      
*
*     Write out performance of the QR factorization to standard output
*
      IF(ICALLER.EQ.0) THEN
         WRITE(*,'(a,e10.3)') 'QR-factorization took ',
     $                        (END_TIME-START_TIME)
         WRITE(*,*) 'seconds. In Mflops/sec: '
         WRITE(*,'(e10.3)') (2D0*N**2*M-2*N**3/3D0)/
     $               ((END_TIME-START_TIME)*1D6)
      END IF
*
*     Apply the matrix Q^T onto B by using the ScaLAPACK routine PDORMQR
*
*     TODO      

*
*     Compute X as X <- inv(R)*B1, X will overwrite B1.
*     X will then be the least-square solution to the problem AX=B. Notice that
*     R is an upper-triangular NxN matrix. Use the PBLAS routine PDTRSM.
*     
*     TODO
      
*      
*     Find now out how accurate the least-square solution is! First
*     do the PDGEMM operation B <- AX - B...
*
*     TODO     

*
*     ...and then take the Frobenius norm of the matrix B
*      
*     TODO

*
      IF(ICALLER.EQ.0) THEN
         WRITE(*,'(a,e10.3)') 'The Frobenius norm of AX-B is ', BNORM
      END IF
*      
*     Release the process grid
*      
      CALL BLACS_GRIDEXIT(CONTXT)
*      
*     Terminate BLACS and end program
*
 500  CONTINUE
      CALL BLACS_EXIT(0)
*
      END
*
*     End of program 
*
***********************************************************************
*     Internal subroutines below                                      *
***********************************************************************
*      
      SUBROUTINE PDFROB( A, DESCA, AFNORM )
*
*     This subroutine computes the Frobenius norm of a matrix A
*
      IMPLICIT NONE
*
*     External functions
*
      INTEGER NUMROC
      EXTERNAL NUMROC
*
*     Variable declaration 
*
      DOUBLE PRECISION A(*)
      DOUBLE PRECISION AFNORM
      INTEGER DESCA(*)
*
      INTEGER M, N, MYROW, MYCOL, NPROW, NPCOL, RSRC, CSRC
      INTEGER MB, NB, I, J, LLD, MLOC, NLOC, CONTXT
      DOUBLE PRECISION LOCALSUM
*
      CONTXT   = DESCA(2)
      M        = DESCA(3)
      N        = DESCA(4)
      MB       = DESCA(5)
      NB       = DESCA(6)
      RSRC     = DESCA(7)
      CSRC     = DESCA(8)
      LLD      = DESCA(9)
      call blacs_gridinfo( CONTXT, NPROW, NPCOL, MYROW, MYCOL )
      MLOC     = numroc( M, MB, MYROW, RSRC, NPROW )
      NLOC     = numroc( N, NB, MYCOL, CSRC, NPCOL )
      LOCALSUM = 0.0D+0
      
*     Every process manipulates their matrix elements with a power to 2      
*     and sum up 
*
      DO 10 J = 1, NLOC
         DO 20 I = 1, MLOC
            LOCALSUM = LOCALSUM + A( (J-1)*LLD + I )**2
 20      CONTINUE     
 10   CONTINUE
*
*     Do global reduction operation: Here we sum up at process 0 in 
*     its local LOCALSUM
*      
*     TODO
     
* 
*     Now we let process 0 take the square root and broadcast
*     the result to all processors which receives the result
*      
      IF( MYROW.EQ.0 .AND. MYCOL.EQ.0 ) THEN
*     TODO         
      ELSE
*     TODO         
      END IF
*
      END
*
***********************************************************************
*      
      SUBROUTINE MATINIT( CONTXT, A, DESCA, B, DESCB, MA, NA, NB)
*
*     This subroutine intitializes the two distributed random matrices
*     A and B with the help of the function DRAND.
*
      IMPLICIT NONE
*
      DOUBLE PRECISION A(*), B(*)
      INTEGER DESCA(*), DESCB(*)
      INTEGER CONTXT, MA, NA, NB
*
      INTEGER I, J, LLDA, LLDB, ICALLER, NPROW, NPCOL, MYROW, MYCOL,
     $        BLACS_PNUM
      DOUBLE PRECISION DRAND
      EXTERNAL DRAND, DRANDINIT, BLACS_GRIDINFO, BLACS_PNUM
*
      LLDA = DESCA(9)
      LLDB = DESCB(9)
*
*     Preparing the random generator
*
      CALL BLACS_GRIDINFO( CONTXT, NPROW, NPCOL, MYROW, MYCOL )
      ICALLER = BLACS_PNUM( CONTXT, MYROW, MYCOL )
      CALL DRANDINIT( ICALLER )
*      
*     Assign values to A with multple calls to DRAND()..
*
*     TODO

*      
*     ..and do the same for B
*      
*     TODO      
      
*
      END
*
***********************************************************************
